@@ -1,16 +1,17 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
 import sys
+from collections.abc import Sequence
 from functools import lru_cache
-from typing import Any, Dict, List, Optional, Sequence, Tuple
+from typing import Any
 
 from kitty.cli_stub import HintsCLIOptions
 from kitty.clipboard import set_clipboard_string, set_primary_selection
 from kitty.constants import website_url
 from kitty.fast_data_types import get_options
-from kitty.typing import BossType
-from kitty.utils import resolve_custom_file
+from kitty.typing_compat import BossType, WindowType
+from kitty.utils import get_editor, resolve_custom_file
 
 from ..tui.handler import result_handler
 
@@ -37,7 +38,7 @@ class Mark:
             text: str,
             groupdict: Any,
             is_hyperlink: bool = False,
-            group_id: Optional[str] = None
+            group_id: str | None = None
     ):
         self.index, self.start, self.end = index, start, end
         self.text = text
@@ -45,7 +46,7 @@ class Mark:
         self.is_hyperlink = is_hyperlink
         self.group_id = group_id
 
-    def as_dict(self) -> Dict[str, Any]:
+    def as_dict(self) -> dict[str, Any]:
         return {
             'index': self.index, 'start': self.start, 'end': self.end,
             'text': self.text, 'groupdict': {str(k):v for k, v in (self.groupdict or {}).items()},
@@ -53,7 +54,7 @@ class Mark:
         }
 
 
-def parse_hints_args(args: List[str]) -> Tuple[HintsCLIOptions, List[str]]:
+def parse_hints_args(args: list[str]) -> tuple[HintsCLIOptions, list[str]]:
     from kitty.cli import parse_args
     return parse_args(args, OPTIONS, usage, help_text, 'kitty +kitten hints', result_class=HintsCLIOptions)
 
@@ -116,7 +117,7 @@ controls where to display the selected error message, other options are ignored.
 default={default_regex}
 The regular expression to use when option :option:`--type` is set to
 :code:`regex`, in Perl 5 syntax. If you specify a numbered group in the regular
-expression, only the group will be matched. This allow you to match text
+expression, only the group will be matched. This allows you to match text
 ignoring a prefix/suffix, as needed. The default expression matches lines. To
 match text over multiple lines, things get a little tricky, as line endings
 are a sequence of zero or more null bytes followed by either a carriage return
@@ -148,6 +149,13 @@ Comma separated list of recognized URL prefixes. Defaults to the list of
 prefixes defined by the :opt:`url_prefixes` option in :file:`kitty.conf`.
 
 
+--url-excluded-characters
+default=default
+Characters to exclude when matching URLs. Defaults to the list of characters
+defined by the :opt:`url_excluded_characters` option in :file:`kitty.conf`.
+The syntax for this option is the same as for :opt:`url_excluded_characters`.
+
+
 --word-characters
 Characters to consider as part of a word. In addition, all characters marked as
 alphanumeric in the Unicode database will be considered as word characters.
@@ -208,19 +216,25 @@ bottom.
 --hints-foreground-color
 default=black
 type=str
-The foreground color for hints.
+The foreground color for hints. You can use color names or hex values. For the eight basic
+named terminal colors you can also use the :code:`bright-` prefix to get the bright variant of the
+color.
 
 
 --hints-background-color
 default=green
 type=str
-The background color for hints.
+The background color for hints. You can use color names or hex values. For the eight basic
+named terminal colors you can also use the :code:`bright-` prefix to get the bright variant of the
+color.
 
 
 --hints-text-color
-default=gray
+default=auto
 type=str
-The foreground color for text pointed to by the hints.
+The foreground color for text pointed to by the hints. You can use color names or hex values. For the eight basic
+named terminal colors you can also use the :code:`bright-` prefix to get the bright variant of the
+color. The default is to pick a suitable color automatically.
 
 
 --customize-processing
@@ -242,11 +256,11 @@ help_text = 'Select text from the screen using the keyboard. Defaults to searchi
 usage = ''
 
 
-def main(args: List[str]) -> Optional[Dict[str, Any]]:
+def main(args: list[str]) -> dict[str, Any] | None:
     raise SystemExit('Should be run as kitten hints')
 
 
-def linenum_process_result(data: Dict[str, Any]) -> Tuple[str, int]:
+def linenum_process_result(data: dict[str, Any]) -> tuple[str, int]:
     for match, g in zip(data['match'], data['groupdicts']):
         path, line = g['path'], g['line']
         if path and line:
@@ -254,12 +268,15 @@ def linenum_process_result(data: Dict[str, Any]) -> Tuple[str, int]:
     return '', -1
 
 
-def linenum_handle_result(args: List[str], data: Dict[str, Any], target_window_id: int, boss: BossType, extra_cli_args: Sequence[str], *a: Any) -> None:
+def linenum_handle_result(args: list[str], data: dict[str, Any], target_window_id: int, boss: BossType, extra_cli_args: Sequence[str], *a: Any) -> None:
     path, line = linenum_process_result(data)
     if not path:
         return
 
-    cmd = [x.format(path=path, line=line) for x in extra_cli_args or ('vim', '+{line}', '{path}')]
+    if extra_cli_args:
+        cmd = [x.format(path=path, line=line) for x in extra_cli_args]
+    else:
+        cmd = get_editor(path_to_edit=path, line_number=line)
     w = boss.window_id_map.get(target_window_id)
     action = data['linenum_action']
 
@@ -279,13 +296,15 @@ def linenum_handle_result(args: List[str], data: Dict[str, Any], target_window_i
                         w.paste_bytes(text)
                     elif program == '@':
                         set_clipboard_string(text)
+                        boss.handle_clipboard_loss('clipboard')
                     elif program == '*':
                         set_primary_selection(text)
+                        boss.handle_clipboard_loss('primary')
                     elif program.startswith('@'):
                         boss.set_clipboard_buffer(program[1:], text)
             else:
                 import shlex
-                text = ' '.join(shlex.quote(arg) for arg in cmd)
+                text = shlex.join(cmd)
                 w.paste_bytes(f'{text}\r')
     elif action == 'background':
         import subprocess
@@ -296,8 +315,15 @@ def linenum_handle_result(args: List[str], data: Dict[str, Any], target_window_i
             }[action])(*cmd)
 
 
-@result_handler(type_of_input='screen-ansi', has_ready_notification=True)
-def handle_result(args: List[str], data: Dict[str, Any], target_window_id: int, boss: BossType) -> None:
+def on_mark_clicked(boss: BossType, window: WindowType, url: str, hyperlink_id: int, cwd: str) -> bool:
+    if url.startswith('mark:'):
+        window.send_cmd_response({'Type': 'mark_activated', 'Mark': int(url[5:])})
+        return True
+    return False
+
+
+@result_handler(type_of_input='screen-ansi', has_ready_notification=True, open_url_handler=on_mark_clicked)
+def handle_result(args: list[str], data: dict[str, Any], target_window_id: int, boss: BossType) -> None:
     cp = data['customize_processing']
     if data['type'] == 'linenum':
         cp = '::linenum::'
@@ -308,7 +334,7 @@ def handle_result(args: List[str], data: Dict[str, Any], target_window_id: int,
             return None
 
     programs = data['programs'] or ('default',)
-    matches: List[str] = []
+    matches: list[str] = []
     groupdicts = []
     for m, g in zip(data['match'], data['groupdicts']):
         if m:
@@ -316,12 +342,12 @@ def handle_result(args: List[str], data: Dict[str, Any], target_window_id: int,
             groupdicts.append(g)
     joiner = data['multiple_joiner']
     try:
-        is_int: Optional[int] = int(joiner)
+        is_int: int | None = int(joiner)
     except Exception:
         is_int = None
     text_type = data['type']
 
-    @lru_cache()
+    @lru_cache
     def joined_text() -> str:
         if is_int is not None:
             try:
