@@ -4,8 +4,9 @@
 
 import os
 import shutil
+from collections.abc import Container, Iterable, Iterator, Sequence
 from contextlib import suppress
-from typing import Any, Container, Dict, FrozenSet, Iterable, Iterator, List, NamedTuple, Optional, Sequence, Tuple
+from typing import Any, Callable, NamedTuple, TypedDict
 
 from .boss import Boss
 from .child import Child
@@ -15,50 +16,84 @@ from .clipboard import set_clipboard_string, set_primary_selection
 from .fast_data_types import add_timer, get_boss, get_options, get_os_window_title, patch_color_profiles
 from .options.utils import env as parse_env
 from .tabs import Tab, TabManager
-from .types import OverlayType, run_once
-from .utils import cmdline_for_hold, get_editor, log_error, resolve_custom_file, resolved_shell, which
+from .types import LayerShellConfig, OverlayType, run_once
+from .utils import get_editor, log_error, resolve_custom_file, which
 from .window import CwdRequest, CwdRequestType, Watchers, Window
 
-try:
-    from typing import TypedDict
-except ImportError:
-    TypedDict = dict
-
 
 class LaunchSpec(NamedTuple):
     opts: LaunchCLIOptions
-    args: List[str]
+    args: list[str]
+
+
+env_docs = '''\
+type=list
+Environment variables to set in the child process. Can be specified multiple
+times to set different environment variables. Syntax: :code:`name=value`. Using
+:code:`name=` will set to empty string and just :code:`name` will remove the
+environment variable.
+'''
+
+
+remote_control_password_docs = '''\
+type=list
+Restrict the actions remote control is allowed to take. This works like
+:opt:`remote_control_password`. You can specify a password and list of actions
+just as for :opt:`remote_control_password`. For example::
+
+    --remote-control-password '"my passphrase" get-* set-colors'
+
+This password will be in effect for this window only.
+Note that any passwords you have defined for :opt:`remote_control_password`
+in :file:`kitty.conf` are also in effect. You can override them by using the same password here.
+You can also disable all :opt:`remote_control_password` global passwords for this window, by using::
+
+    --remote-control-password '!'
+
+This option only takes effect if :option:`--allow-remote-control`
+is also specified. Can be specified multiple times to create multiple passwords.
+This option was added to kitty in version 0.26.0
+'''
 
 
 @run_once
 def options_spec() -> str:
-    return '''
+    return f"""
+--source-window
+A match expression to select the window from which data such as title, colors, env vars,
+screen contents, etc. are copied. When not specified the currently active window is used.
+See :ref:`search_syntax` for the syntax used for specifying windows.
+
+
 --window-title --title
 The title to set for the new window. By default, title is controlled by the
 child process. The special value :code:`current` will copy the title from the
-currently active window.
+:option:`source window <launch --source-window>`.
 
 
 --tab-title
 The title for the new tab if launching in a new tab. By default, the title
 of the active window in the tab is used as the tab title. The special value
-:code:`current` will copy the title from the title of the currently active tab.
+:code:`current` will copy the title from the tab containing the
+:option:`source window <launch --source-window>`.
 
 
 --type
 type=choices
 default=window
-choices=window,tab,os-window,overlay,overlay-main,background,clipboard,primary
+choices=window,tab,os-window,os-panel,overlay,overlay-main,background,clipboard,primary
 Where to launch the child process:
 
 :code:`window`
     A new :term:`kitty window <window>` in the current tab
 
 :code:`tab`
-    A new :term:`tab` in the current OS window
+    A new :term:`tab` in the current OS window. Not available when the
+    :doc:`launch <launch>` command is used in :ref:`startup sessions <sessions>`.
 
 :code:`os-window`
-    A new :term:`operating system window <os_window>`
+    A new :term:`operating system window <os_window>`.  Not available when the
+    :doc:`launch <launch>` command is used in :ref:`startup sessions <sessions>`.
 
 :code:`overlay`
     An :term:`overlay window <overlay>` covering the current active kitty window
@@ -81,6 +116,11 @@ Where to launch the child process:
     These two are meant to work with :option:`--stdin-source <launch --stdin-source>` to copy
     data to the :italic:`system clipboard` or :italic:`primary selection`.
 
+:code:`os-panel`
+    Similar to :code:`os-window`, except that it creates the new OS Window as a desktop panel.
+    Only works on platforms that support this, such as Wayand compositors that support the layer
+    shell protocol. Use the :option:`kitten @ launch --os-panel` option to configure the panel.
+
 #placeholder_for_formatting#
 
 
@@ -93,7 +133,7 @@ opened window.
 --cwd
 completion=type:directory kwds:current,oldest,last_reported,root
 The working directory for the newly launched child. Use the special value
-:code:`current` to use the working directory of the currently active window.
+:code:`current` to use the working directory of the :option:`source window <launch --source-window>`
 The special value :code:`last_reported` uses the last working directory reported
 by the shell (needs :ref:`shell_integration` to work). The special value
 :code:`oldest` works like :code:`current` but uses the working directory of the
@@ -101,13 +141,13 @@ oldest foreground process associated with the currently active window rather
 than the newest foreground process. Finally, the special value :code:`root`
 refers to the process that was originally started when the window was created.
 
+When opening in the same working directory as the current window causes the new
+window to connect to a remote host, you can use the :option:`--hold-after-ssh`
+flag to prevent the new window from closing when the connection is terminated.
+
 
 --env
-type=list
-Environment variables to set in the child process. Can be specified multiple
-times to set different environment variables. Syntax: :code:`name=value`. Using
-:code:`name=` will set to empty string and just :code:`name` will remove the
-environment variable.
+{env_docs}
 
 
 --var
@@ -120,23 +160,24 @@ times to set different user variables. Syntax: :code:`name=value`. Using
 --hold
 type=bool-set
 Keep the window open even after the command being executed exits, at a shell prompt.
+The shell will be run after the launched command exits.
 
 
 --copy-colors
 type=bool-set
 Set the colors of the newly created window to be the same as the colors in the
-currently active window.
+:option:`source window <launch --source-window>`.
 
 
 --copy-cmdline
 type=bool-set
 Ignore any specified command line and instead use the command line from the
-currently active window.
+:option:`source window <launch --source-window>`.
 
 
 --copy-env
 type=bool-set
-Copy the environment variables from the currently active window into the newly
+Copy the environment variables from the :option:`source window <launch --source-window>` into the newly
 launched child process. Note that this only copies the environment when the
 window was first created, as it is not possible to get updated environment variables
 from arbitrary processes. To copy that environment, use either the :ref:`clone-in-kitty
@@ -156,7 +197,56 @@ instead of at the end. The values of :code:`vsplit`, :code:`hsplit` and
 :code:`split` are only used by the :code:`splits` layout and control if the new
 window is placed in a vertical, horizontal or automatic split with the currently
 active window. The default is to place the window in a layout dependent manner,
-typically, after the currently active window.
+typically, after the currently active window. See :option:`--next-to <launch --next-to>`
+to use a window other than the currently active window.
+
+
+--next-to
+A match expression to select the window next to which the new window is created.
+See :ref:`search_syntax` for the syntax for specifying windows. If not specified
+defaults to the active window. When used via remote control and a target tab is
+specified this option is ignored unless the matched window is in the specified tab.
+When using :option:`--type <launch --type>` of :code:`tab`, the tab will be created
+in the OS Window containing the matched window.
+
+
+--bias
+type=float
+default=0
+The bias used to alter the size of the window.
+It controls what fraction of available space the window takes. The exact meaning
+of bias depends on the current layout.
+
+* Splits layout: The bias is interpreted as a percentage between 0 and 100.
+When splitting a window into two, the new window will take up the specified fraction
+of the space allotted to the original window and the original window will take up
+the remainder of the space.
+
+* Vertical/horizontal layout: The bias is interpreted as adding/subtracting from the
+normal size of the window. It should be a number between -90 and 90. This number is
+the percentage of the OS Window size that should be added to the window size.
+So for example, if a window would normally have been size 50 in the layout inside an
+OS Window that is size 80 high and --bias -10 is used it will become *approximately*
+size 42 high. Note that sizes are approximations, you cannot use this method to
+create windows of fixed sizes.
+
+* Tall layout: If the window being created is the *first* window in a column, then
+the bias is interpreted as a percentage, as for the splits layout, splitting the OS
+Window width between columns. If the window is a second or subsequent window in a column
+the bias is interpreted as adding/subtracting from the window size as for the vertical
+layout above.
+
+* Fat layout: Same as tall layout except it goes by rows instead of columns.
+
+* Grid layout: The bias is interpreted the same way as for the Vertical and Horizontal
+layouts, as something to be added/subtracted to the normal size. However, the
+since in a grid layout there are rows *and* columns, the bias on the first window in a column
+operates on the columns. Any later windows in that column operate on the row.
+So, for example, if you bias the first window in a grid layout it will change the width
+of the first column, the second window, the width of the second column, the third window,
+the height of the second row and so on.
+
+The bias option was introduced in kitty version 0.36.0.
 
 
 --allow-remote-control
@@ -171,24 +261,7 @@ remote control.
 
 
 --remote-control-password
-type=list
-Restrict the actions remote control is allowed to take. This works like
-:opt:`remote_control_password`. You can specify a password and list of actions
-just as for :opt:`remote_control_password`. For example::
-
-    --remote-control-password '"my passphrase" get-* set-colors'
-
-This password will be in effect for this window only.
-Note that any passwords you have defined for :opt:`remote_control_password`
-in :file:`kitty.conf` are also in effect. You can override them by using the same password here.
-You can also disable all :opt:`remote_control_password` global passwords for this window, by using::
-
-    --remote-control-password '!'
-
-This option only takes effect if :option:`--allow-remote-control`
-is also specified. Can be specified multiple times to create multiple passwords.
-This option was added to kitty in version 0.26.0
-
+{remote_control_password_docs}
 
 --stdin-source
 type=choices
@@ -197,16 +270,16 @@ choices=none,@selection,@screen,@screen_scrollback,@alternate,@alternate_scrollb
 Pass the screen contents as :file:`STDIN` to the child process.
 
 :code:`@selection`
-    is the currently selected text.
+    is the currently selected text in the :option:`source window <launch --source-window>`.
 
 :code:`@screen`
-    is the contents of the currently active window.
+    is the contents of the :option:`source window <launch --source-window>`.
 
 :code:`@screen_scrollback`
     is the same as :code:`@screen`, but includes the scrollback buffer as well.
 
 :code:`@alternate`
-    is the secondary screen of the current active window. For example if you run
+    is the secondary screen of the :option:`source window <launch --source-window>`. For example if you run
     a full screen terminal application, the secondary screen will
     be the screen you return to when quitting the application.
 
@@ -258,7 +331,8 @@ using :option:`--type=os-window <launch --type>`. Defaults to
 --os-window-title
 Set the title for the newly created OS window. This title will override any
 titles set by programs running in kitty. The special value :code:`current` will
-use the title of the current OS window, if any.
+copy the title from the OS Window containing the
+:option:`source window <launch --source-window>`.
 
 
 --os-window-state
@@ -313,20 +387,37 @@ the section on watchers in the launch command documentation: :ref:`watchers`.
 Relative paths are resolved relative to the :ref:`kitty config directory
 <confloc>`. Global watchers for all windows can be specified with
 :opt:`watcher` in :file:`kitty.conf`.
-'''
 
 
-def parse_launch_args(args: Optional[Sequence[str]] = None) -> LaunchSpec:
+--os-panel
+type=list
+Options to control the creation of desktop panels. Takes the same settings
+as the :doc:`panel kitten </kittens/panel>`, except for :option:`--override <kitty +kitten panel --override>`
+and :option:`--config <kitty +kitten panel --config>`. Can be specified multiple times.
+For example, to create a desktop panel at the bottom of the screen two lines high::
+
+    launch --type os-panel --os-panel lines=2 --os-panel edge=bottom sh -c "echo; echo; echo hello; sleep 5s"
+
+
+--hold-after-ssh
+type=bool-set
+When using :option:`--cwd`:code:`=current` or similar from a window that is running the ssh kitten,
+the new window will run a local shell after disconnecting from the remote host, when this option
+is specified.
+"""
+
+
+def parse_launch_args(args: Sequence[str] | None = None) -> LaunchSpec:
     args = list(args or ())
     try:
         opts, args = parse_args(result_class=LaunchCLIOptions, args=args, ospec=options_spec)
     except SystemExit as e:
-        raise ValueError from e
+        raise ValueError(str(e)) from e
     return LaunchSpec(opts, args)
 
 
-def get_env(opts: LaunchCLIOptions, active_child: Optional[Child] = None, base_env: Optional[Dict[str,str]] = None) -> Dict[str, str]:
-    env: Dict[str, str] = {}
+def get_env(opts: LaunchCLIOptions, active_child: Child | None = None, base_env: dict[str,str] | None = None) -> dict[str, str]:
+    env: dict[str, str] = {}
     if opts.copy_env and active_child:
         env.update(active_child.foreground_environ)
     if base_env is not None:
@@ -337,15 +428,28 @@ def get_env(opts: LaunchCLIOptions, active_child: Optional[Child] = None, base_e
     return env
 
 
-def tab_for_window(boss: Boss, opts: LaunchCLIOptions, target_tab: Optional[Tab] = None) -> Optional[Tab]:
+def layer_shell_config_from_panel_opts(panel_opts: Iterable[str]) -> LayerShellConfig:
+    from kittens.panel.main import layer_shell_config, parse_panel_args
+    args = [('' if x.startswith('--') else '--') + x for x in panel_opts]
+    try:
+        opts, _ = parse_panel_args(args)
+    except SystemExit as e:
+        raise ValueError(str(e))
+    return layer_shell_config(opts)
 
-    def create_tab(tm: Optional[TabManager] = None) -> Tab:
+
+def tab_for_window(boss: Boss, opts: LaunchCLIOptions, target_tab: Tab | None, next_to: Window | None) -> Tab:
+
+    def create_tab(tm: TabManager | None = None) -> Tab:
         if tm is None:
-            oswid = boss.add_os_window(
-                wclass=opts.os_window_class,
-                wname=opts.os_window_name,
-                window_state=opts.os_window_state,
-                override_title=opts.os_window_title or None)
+            if opts.type == 'os-panel':
+                oswid = boss.add_os_panel(layer_shell_config_from_panel_opts(opts.os_panel), opts.os_window_class, opts.os_window_name)
+            else:
+                oswid = boss.add_os_window(
+                    wclass=opts.os_window_class,
+                    wname=opts.os_window_name,
+                    window_state=opts.os_window_state,
+                    override_title=opts.os_window_title or None)
             tm = boss.os_window_map[oswid]
         tab = tm.new_tab(empty_tab=True, location=opts.location)
         if opts.tab_title:
@@ -353,27 +457,34 @@ def tab_for_window(boss: Boss, opts: LaunchCLIOptions, target_tab: Optional[Tab]
         return tab
 
     if opts.type == 'tab':
+        tm = boss.active_tab_manager
+        if next_to is not None and (tab := next_to.tabref()) is not None and (qtm := tab.tab_manager_ref()) is not None:
+            tm = qtm
         if target_tab is not None:
-            tm = target_tab.tab_manager_ref() or boss.active_tab_manager
-        else:
-            tm = boss.active_tab_manager
+            tm = target_tab.tab_manager_ref() or tm
         tab = create_tab(tm)
-    elif opts.type == 'os-window':
+    elif opts.type in ('os-window', 'os-panel'):
         tab = create_tab()
     else:
-        tab = target_tab or boss.active_tab or create_tab()
-
+        if target_tab is not None:
+            tab = target_tab
+        else:
+            tab = boss.active_tab
+            if next_to is not None and (qtab := next_to.tabref()) is not None:
+                tab = qtab
+            tab = tab or create_tab()
     return tab
 
 
-watcher_modules: Dict[str, Any] = {}
+watcher_modules: dict[str, Any] = {}
 
 
-def load_watch_modules(watchers: Iterable[str]) -> Optional[Watchers]:
+def load_watch_modules(watchers: Iterable[str]) -> Watchers | None:
     if not watchers:
         return None
     import runpy
     ans = Watchers()
+    boss = get_boss()
     for path in watchers:
         path = resolve_custom_file(path)
         m = watcher_modules.get(path, None)
@@ -387,6 +498,14 @@ def load_watch_modules(watchers: Iterable[str]) -> Optional[Watchers]:
                 watcher_modules[path] = False
                 continue
             watcher_modules[path] = m
+            w = m.get('on_load')
+            if callable(w):
+                try:
+                    w(boss, {})
+                except Exception as err:
+                    import traceback
+                    log_error(traceback.format_exc())
+                    log_error(f'Failed to call on_load() in watcher from {path} with error: {err}')
         if m is False:
             continue
         w = m.get('on_close')
@@ -398,32 +517,47 @@ def load_watch_modules(watchers: Iterable[str]) -> Optional[Watchers]:
         w = m.get('on_focus_change')
         if callable(w):
             ans.on_focus_change.append(w)
+        w = m.get('on_set_user_var')
+        if callable(w):
+            ans.on_set_user_var.append(w)
+        w = m.get('on_title_change')
+        if callable(w):
+            ans.on_title_change.append(w)
+        w = m.get('on_cmd_startstop')
+        if callable(w):
+            ans.on_cmd_startstop.append(w)
+        w = m.get('on_color_scheme_preference_change')
+        if callable(w):
+            ans.on_color_scheme_preference_change.append(w)
     return ans
 
 
 class LaunchKwds(TypedDict):
 
     allow_remote_control: bool
-    remote_control_passwords: Optional[Dict[str, Sequence[str]]]
-    cwd_from: Optional[CwdRequest]
-    cwd: Optional[str]
-    location: Optional[str]
-    override_title: Optional[str]
-    copy_colors_from: Optional[Window]
-    marker: Optional[str]
-    cmd: Optional[List[str]]
-    overlay_for: Optional[int]
-    stdin: Optional[bytes]
+    remote_control_passwords: dict[str, Sequence[str]] | None
+    cwd_from: CwdRequest | None
+    cwd: str | None
+    location: str | None
+    override_title: str | None
+    copy_colors_from: Window | None
+    marker: str | None
+    cmd: list[str] | None
+    overlay_for: int | None
+    stdin: bytes | None
+    hold: bool
+    bias: float | None
+    hold_after_ssh: bool
 
 
 def apply_colors(window: Window, spec: Sequence[str]) -> None:
-    from kitty.rc.set_colors import parse_colors
-    colors = parse_colors(spec)
+    from .colors import parse_colors
+    colors, transparent_background_colors = parse_colors(spec)
     profiles = window.screen.color_profile,
-    patch_color_profiles(colors, profiles, True)
+    patch_color_profiles(colors, transparent_background_colors, profiles, True)
 
 
-def parse_var(defn: Iterable[str]) -> Iterator[Tuple[str, str]]:
+def parse_var(defn: Iterable[str]) -> Iterator[tuple[str, str]]:
     for item in defn:
         a, sep, b = item.partition('=')
         yield a, b
@@ -452,41 +586,57 @@ force_window_launch = ForceWindowLaunch()
 non_window_launch_types = 'background', 'clipboard', 'primary'
 
 
+def parse_remote_control_passwords(allow_remote_control: bool, passwords: Sequence[str]) -> dict[str, Sequence[str]] | None:
+    remote_control_restrictions: dict[str, Sequence[str]] | None = None
+    if allow_remote_control and passwords:
+        from kitty.options.utils import remote_control_password
+        remote_control_restrictions = {}
+        for rcp in passwords:
+            for pw, rcp_items in remote_control_password(rcp, {}):
+                remote_control_restrictions[pw] = rcp_items
+    return remote_control_restrictions
+
+
 def _launch(
     boss: Boss,
     opts: LaunchCLIOptions,
-    args: List[str],
-    target_tab: Optional[Tab] = None,
+    args: list[str],
+    target_tab: Tab | None = None,
     force_target_tab: bool = False,
-    active: Optional[Window] = None,
     is_clone_launch: str = '',
-    rc_from_window: Optional[Window] = None,
-    base_env: Optional[Dict[str, str]] = None,
-) -> Optional[Window]:
-    active = active or boss.active_window_for_cwd
-    if active:
-        active_child = active.child
-    else:
-        active_child = None
+    rc_from_window: Window | None = None,
+    base_env: dict[str, str] | None = None,
+    child_death_callback: Callable[[int, Exception | None], None] | None = None,
+) -> Window | None:
+    source_window = boss.active_window_for_cwd
+    if opts.source_window:
+        for qw in boss.match_windows(opts.source_window, rc_from_window):
+            source_window = qw
+            break
+    source_child = source_window.child if source_window else None
+    next_to = boss.active_window
+    if opts.next_to:
+        for qw in boss.match_windows(opts.next_to, rc_from_window):
+            next_to = qw
+            break
+    if target_tab and next_to and next_to not in target_tab:
+        next_to = None
     if opts.window_title == 'current':
-        opts.window_title = active.title if active else None
+        opts.window_title = source_window.title if source_window else None
     if opts.tab_title == 'current':
         atab = boss.active_tab
+        if source_window and (qt := source_window.tabref()):
+            atab = qt
         opts.tab_title = atab.effective_title if atab else None
     if opts.os_window_title == 'current':
         tm = boss.active_tab_manager
+        if source_window and (qt := source_window.tabref()) and (qr := qt.tab_manager_ref()):
+            tm = qr
         opts.os_window_title = get_os_window_title(tm.os_window_id) if tm else None
-    env = get_env(opts, active_child, base_env)
-    remote_control_restrictions: Optional[Dict[str, Sequence[str]]] = None
-    if opts.allow_remote_control and opts.remote_control_password:
-        from kitty.options.utils import remote_control_password
-        remote_control_restrictions = {}
-        for rcp in opts.remote_control_password:
-            for pw, rcp_items in remote_control_password(rcp, {}):
-                remote_control_restrictions[pw] = rcp_items
+    env = get_env(opts, source_child, base_env)
     kw: LaunchKwds = {
         'allow_remote_control': opts.allow_remote_control,
-        'remote_control_passwords': remote_control_restrictions,
+        'remote_control_passwords': parse_remote_control_passwords(opts.allow_remote_control, opts.remote_control_password),
         'cwd_from': None,
         'cwd': None,
         'location': None,
@@ -495,34 +645,42 @@ def _launch(
         'marker': opts.marker or None,
         'cmd': None,
         'overlay_for': None,
-        'stdin': None
+        'stdin': None,
+        'hold': False,
+        'bias': None,
+        'hold_after_ssh': False
     }
     spacing = {}
     if opts.spacing:
         from .rc.set_spacing import parse_spacing_settings, patch_window_edges
         spacing = parse_spacing_settings(opts.spacing)
+    if opts.bias:
+        kw['bias'] = max(-100, min(opts.bias, 100))
     if opts.cwd:
         if opts.cwd == 'current':
-            if active:
-                kw['cwd_from'] = CwdRequest(active)
+            if source_window:
+                kw['cwd_from'] = CwdRequest(source_window)
         elif opts.cwd == 'last_reported':
-            if active:
-                kw['cwd_from'] = CwdRequest(active, CwdRequestType.last_reported)
+            if source_window:
+                kw['cwd_from'] = CwdRequest(source_window, CwdRequestType.last_reported)
         elif opts.cwd == 'oldest':
-            if active:
-                kw['cwd_from'] = CwdRequest(active, CwdRequestType.oldest)
+            if source_window:
+                kw['cwd_from'] = CwdRequest(source_window, CwdRequestType.oldest)
         elif opts.cwd == 'root':
-            if active:
-                kw['cwd_from'] = CwdRequest(active, CwdRequestType.root)
+            if source_window:
+                kw['cwd_from'] = CwdRequest(source_window, CwdRequestType.root)
         else:
             kw['cwd'] = opts.cwd
-        if kw['cwd_from'] is not None and rc_from_window is not None:
-            kw['cwd_from'].rc_from_window_id = rc_from_window.id
+    if opts.hold_after_ssh:
+        if opts.cwd not in ('current', 'last_reported', 'oldest'):
+            raise ValueError("--hold-after-ssh can only be supplied if --cwd=current or similar is also supplied")
+        kw['hold_after_ssh'] = True
+
     if opts.location != 'default':
         kw['location'] = opts.location
-    if opts.copy_colors and active:
-        kw['copy_colors_from'] = active
-    pipe_data: Dict[str, Any] = {}
+    if opts.copy_colors and source_window:
+        kw['copy_colors_from'] = source_window
+    pipe_data: dict[str, Any] = {}
     if opts.stdin_source != 'none':
         q = str(opts.stdin_source)
         if opts.stdin_add_formatting:
@@ -531,25 +689,27 @@ def _launch(
                 q = f'@ansi_{q[1:]}'
         if opts.stdin_add_line_wrap_markers:
             q += '_wrap'
-        penv, stdin = boss.process_stdin_source(window=active, stdin=q, copy_pipe_data=pipe_data)
+        penv, stdin = boss.process_stdin_source(window=source_window, stdin=q, copy_pipe_data=pipe_data)
         if stdin:
             kw['stdin'] = stdin
             if penv:
                 env.update(penv)
 
     cmd = args or None
-    if opts.copy_cmdline and active_child:
-        cmd = active_child.foreground_cmdline
+    if opts.copy_cmdline and source_child:
+        cmd = source_child.foreground_cmdline
+    active = boss.active_window
     if cmd:
-        final_cmd: List[str] = []
+        final_cmd: list[str] = []
         for x in cmd:
-            if active and not opts.copy_cmdline:
+            if source_window and not opts.copy_cmdline:
                 if x == '@selection':
-                    s = boss.data_for_at(which=x, window=active)
+                    s = boss.data_for_at(which=x, window=source_window)
                     if s:
                         x = s
                 elif x == '@active-kitty-window-id':
-                    x = str(active.id)
+                    if active:
+                        x = str(active.id)
                 elif x == '@input-line-number':
                     if 'input_line_number' in pipe_data:
                         x = str(pipe_data['input_line_number'])
@@ -557,8 +717,8 @@ def _launch(
                     if 'lines' in pipe_data:
                         x = str(pipe_data['lines'])
                 elif x in ('@cursor-x', '@cursor-y', '@scrolled-by', '@first-line-on-screen', '@last-line-on-screen'):
-                    if active is not None:
-                        screen = active.screen
+                    if source_window is not None:
+                        screen = source_window.screen
                         if x == '@scrolled-by':
                             x = str(screen.scrolled_by)
                         elif x == '@cursor-x':
@@ -577,8 +737,10 @@ def _launch(
         kw['cmd'] = final_cmd
     if force_window_launch and opts.type not in non_window_launch_types:
         opts.type = 'window'
-    base_for_overlay = active
-    if target_tab:
+    if next_to and opts.type in non_window_launch_types:
+        next_to = None
+    base_for_overlay = next_to
+    if target_tab and (not base_for_overlay or base_for_overlay not in target_tab):
         base_for_overlay = target_tab.active_window
     if opts.type in ('overlay', 'overlay-main') and base_for_overlay:
         kw['overlay_for'] = base_for_overlay.id
@@ -588,81 +750,83 @@ def _launch(
             raise ValueError('The cmd to run must be specified when running a background process')
         boss.run_background_process(
             cmd, cwd=kw['cwd'], cwd_from=kw['cwd_from'], env=env or None, stdin=kw['stdin'],
-            allow_remote_control=kw['allow_remote_control'], remote_control_passwords=kw['remote_control_passwords']
+            allow_remote_control=kw['allow_remote_control'], remote_control_passwords=kw['remote_control_passwords'],
+            notify_on_death=child_death_callback,
         )
     elif opts.type in ('clipboard', 'primary'):
         stdin = kw.get('stdin')
         if stdin is not None:
             if opts.type == 'clipboard':
                 set_clipboard_string(stdin)
+                boss.handle_clipboard_loss('clipboard')
             else:
                 set_primary_selection(stdin)
+                boss.handle_clipboard_loss('primary')
+        if child_death_callback is not None:
+            child_death_callback(0, None)
     else:
-        if opts.hold:
-            cmd = kw['cmd'] or resolved_shell()
-            if not os.path.isabs(cmd[0]):
-                cmd[0] = which(cmd[0]) or cmd[0]
-            kw['cmd'] = cmdline_for_hold(cmd)
-        if force_target_tab:
+        kw['hold'] = opts.hold
+        if force_target_tab and target_tab is not None:
             tab = target_tab
         else:
-            tab = tab_for_window(boss, opts, target_tab)
-        if tab is not None:
-            watchers = load_watch_modules(opts.watcher)
-            with Window.set_ignore_focus_changes_for_new_windows(opts.keep_focus):
-                new_window: Window = tab.new_window(
-                    env=env or None, watchers=watchers or None, is_clone_launch=is_clone_launch, **kw)
-            if spacing:
-                patch_window_edges(new_window, spacing)
-                tab.relayout()
-            if opts.color:
-                apply_colors(new_window, opts.color)
-            if opts.keep_focus:
-                if active:
-                    boss.set_active_window(active, switch_os_window_if_needed=True, for_keep_focus=True)
-                if not Window.initial_ignore_focus_changes_context_manager_in_operation:
-                    new_window.ignore_focus_changes = False
-            if opts.logo:
-                new_window.set_logo(opts.logo, opts.logo_position or '', opts.logo_alpha)
-            if opts.type == 'overlay-main':
-                new_window.overlay_type = OverlayType.main
-            if opts.var:
-                for key, val in parse_var(opts.var):
-                    new_window.set_user_var(key, val)
-            return new_window
+            tab = tab_for_window(boss, opts, target_tab, next_to)
+        watchers = load_watch_modules(opts.watcher)
+        with Window.set_ignore_focus_changes_for_new_windows(opts.keep_focus):
+            new_window: Window = tab.new_window(
+                env=env or None, watchers=watchers or None, is_clone_launch=is_clone_launch, next_to=next_to, **kw)
+            if child_death_callback is not None:
+                boss.monitor_pid(new_window.child.pid or 0, child_death_callback)
+        if spacing:
+            patch_window_edges(new_window, spacing)
+            tab.relayout()
+        if opts.color:
+            apply_colors(new_window, opts.color)
+        if opts.keep_focus:
+            if active:
+                boss.set_active_window(active, switch_os_window_if_needed=True, for_keep_focus=True)
+            if not Window.initial_ignore_focus_changes_context_manager_in_operation:
+                new_window.ignore_focus_changes = False
+        if opts.logo:
+            new_window.set_logo(opts.logo, opts.logo_position or '', opts.logo_alpha)
+        if opts.type == 'overlay-main':
+            new_window.overlay_type = OverlayType.main
+        if opts.var:
+            for key, val in parse_var(opts.var):
+                new_window.set_user_var(key, val)
+        return new_window
     return None
 
 
 def launch(
     boss: Boss,
     opts: LaunchCLIOptions,
-    args: List[str],
-    target_tab: Optional[Tab] = None,
+    args: list[str],
+    target_tab: Tab | None = None,
     force_target_tab: bool = False,
-    active: Optional[Window] = None,
     is_clone_launch: str = '',
-    rc_from_window: Optional[Window] = None,
-    base_env: Optional[Dict[str, str]] = None,
-) -> Optional[Window]:
-    active = active or boss.active_window_for_cwd
+    rc_from_window: Window | None = None,
+    base_env: dict[str, str] | None = None,
+    child_death_callback: Callable[[int, Exception | None], None] | None = None,
+) -> Window | None:
+    active = boss.active_window
     if opts.keep_focus and active:
         orig, active.ignore_focus_changes = active.ignore_focus_changes, True
     try:
-        return _launch(boss, opts, args, target_tab, force_target_tab, active, is_clone_launch, rc_from_window, base_env)
+        return _launch(boss, opts, args, target_tab, force_target_tab, is_clone_launch, rc_from_window, base_env, child_death_callback)
     finally:
         if opts.keep_focus and active:
             active.ignore_focus_changes = orig
 
 @run_once
-def clone_safe_opts() -> FrozenSet[str]:
+def clone_safe_opts() -> frozenset[str]:
     return frozenset((
         'window_title', 'tab_title', 'type', 'keep_focus', 'cwd', 'env', 'var', 'hold',
         'location', 'os_window_class', 'os_window_name', 'os_window_title', 'os_window_state',
-        'logo', 'logo_position', 'logo_alpha', 'color', 'spacing',
+        'logo', 'logo_position', 'logo_alpha', 'color', 'spacing', 'next_to', 'hold_after_ssh'
     ))
 
 
-def parse_opts_for_clone(args: List[str]) -> Tuple[LaunchCLIOptions, List[str]]:
+def parse_opts_for_clone(args: list[str]) -> tuple[LaunchCLIOptions, list[str]]:
     unsafe, unsafe_args = parse_launch_args(args)
     default_opts, default_args = parse_launch_args()
     # only copy safe options, those that dont lead to local code exec
@@ -671,7 +835,7 @@ def parse_opts_for_clone(args: List[str]) -> Tuple[LaunchCLIOptions, List[str]]:
     return default_opts, unsafe_args
 
 
-def parse_null_env(text: str) -> Dict[str, str]:
+def parse_null_env(text: str) -> dict[str, str]:
     ans = {}
     for line in text.split('\0'):
         if line:
@@ -683,7 +847,7 @@ def parse_null_env(text: str) -> Dict[str, str]:
     return ans
 
 
-def parse_message(msg: str, simple: Container[str]) -> Iterator[Tuple[str, str]]:
+def parse_message(msg: str, simple: Container[str]) -> Iterator[tuple[str, str]]:
     from base64 import standard_b64decode
     for x in msg.split(','):
         try:
@@ -699,7 +863,7 @@ class EditCmd:
 
     def __init__(self, msg: str) -> None:
         self.tdir = ''
-        self.args: List[str] = []
+        self.args: list[str] = []
         self.cwd = self.file_name = self.file_localpath = ''
         self.file_data = b''
         self.file_inode = -1, -1
@@ -773,7 +937,7 @@ class EditCmd:
     def on_edit_window_close(self, window: Window) -> None:
         self.check_status()
 
-    def check_status(self, timer_id: Optional[int] = None) -> None:
+    def check_status(self, timer_id: int | None = None) -> None:
         if self.abort_signaled:
             return
         boss = get_boss()
@@ -808,8 +972,8 @@ class EditCmd:
 class CloneCmd:
 
     def __init__(self, msg: str) -> None:
-        self.args: List[str] = []
-        self.env: Optional[Dict[str, str]] = None
+        self.args: list[str] = []
+        self.env: dict[str, str] | None = None
         self.cwd = ''
         self.shell = ''
         self.envfmt = 'default'
@@ -844,6 +1008,23 @@ class CloneCmd:
                     # skip SSH environment variables
                     'SSH_CLIENT', 'SSH_CONNECTION', 'SSH_ORIGINAL_COMMAND', 'SSH_TTY', 'SSH2_TTY',
                     'SSH_TUNNEL', 'SSH_USER_AUTH', 'SSH_AUTH_SOCK',
+                    # Dont clone KITTY_WINDOW_ID
+                    'KITTY_WINDOW_ID',
+                    # Bash variables from "bind -x" and "complete -C" (needed not to confuse bash-preexec)
+                    'READLINE_ARGUMENT', 'READLINE_LINE', 'READLINE_MARK', 'READLINE_POINT',
+                    'COMP_LINE', 'COMP_POINT', 'COMP_TYPE',
+                    # GPG gpg-agent
+                    'GPG_TTY',
+                    # Session variables of XDG
+                    'XDG_SESSION_CLASS', 'XDG_SESSION_ID', 'XDG_SESSION_TYPE',
+                    # Session variables of GNU Screen
+                    'STY', 'WINDOW',
+                    # Session variables of interactive shell plugins
+                    'ATUIN_SESSION', 'ATUIN_HISTORY_ID',
+                    'BLE_SESSION_ID', '_ble_util_fdlist_cloexec', '_ble_util_fdvars_export',
+                    '_GITSTATUS_CLIENT_PID', '_GITSTATUS_REQ_FD', '_GITSTATUS_RESP_FD', 'GITSTATUS_DAEMON_PID',
+                    'MCFLY_SESSION_ID',
+                    'STARSHIP_SESSION_KEY',
                 } and not k.startswith((
                     # conda state env vars for multi-level virtual environments
                     'CONDA_PREFIX_',
@@ -854,7 +1035,7 @@ class CloneCmd:
                 self.history = v
 
 
-edits_in_flight: Dict[int, EditCmd] = {}
+edits_in_flight: dict[int, EditCmd] = {}
 
 
 def remote_edit(msg: str, window: Window) -> None:
@@ -865,7 +1046,8 @@ def remote_edit(msg: str, window: Window) -> None:
             q.abort_signaled = c.abort_signaled
         return
     cmdline = get_editor(path_to_edit=c.file_localpath, line_number=c.line_number)
-    w = launch(get_boss(), c.opts, cmdline, active=window)
+    c.opts.source_window = c.opts.next_to = f'id:{window.id}'
+    w = launch(get_boss(), c.opts, cmdline)
     if w is not None:
         c.source_window_id = window.id
         c.editor_window_id = w.id
@@ -926,4 +1108,6 @@ def clone_and_launch(msg: str, window: Window) -> None:
             cmdline[0] = window.child.final_exe
         if cmdline and cmdline == [window.child.final_exe] + window.child.argv[1:]:
             cmdline = window.child.unmodified_argv
-    launch(get_boss(), c.opts, cmdline, active=window, is_clone_launch=is_clone_launch)
+    c.opts.source_window = f'id:{window.id}'
+    c.opts.next_to = c.opts.next_to or c.opts.source_window
+    launch(get_boss(), c.opts, cmdline, is_clone_launch=is_clone_launch)
