@@ -1,16 +1,27 @@
 #!/usr/bin/env python
 # License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>
 
+from collections.abc import Generator, Sequence
 from itertools import islice, repeat
-from typing import Any, Dict, Generator, List, Optional, Sequence, Tuple
+from typing import Any
 
 from kitty.borders import BorderColor
 from kitty.conf.utils import to_bool
 from kitty.types import Edges
-from kitty.typing import EdgeLiteral, WindowType
+from kitty.typing_compat import EdgeLiteral, WindowType
 from kitty.window_list import WindowGroup, WindowList
 
-from .base import BorderLine, Layout, LayoutData, LayoutDimension, LayoutOpts, NeighborsMap, lgd, normalize_biases, safe_increment_bias
+from .base import (
+    BorderLine,
+    Layout,
+    LayoutData,
+    LayoutDimension,
+    LayoutOpts,
+    NeighborsMap,
+    lgd,
+    normalize_biases,
+    safe_increment_bias,
+)
 from .vertical import borders
 
 
@@ -55,7 +66,7 @@ class TallLayoutOpts(LayoutOpts):
     full_size: int = 1
     mirrored: bool = False
 
-    def __init__(self, data: Dict[str, str]):
+    def __init__(self, data: dict[str, str]):
         try:
             self.full_size = int(data.get('full_size', 1))
         except Exception:
@@ -67,15 +78,27 @@ class TallLayoutOpts(LayoutOpts):
             self.bias = 50
         self.mirrored = to_bool(data.get('mirrored', 'false'))
 
-    def serialized(self) -> Dict[str, Any]:
+    def serialized(self) -> dict[str, Any]:
         return {'full_size': self.full_size, 'bias': self.bias, 'mirrored': self.mirrored}
 
-    def build_bias_list(self) -> Tuple[float, ...]:
+    def build_bias_list(self) -> tuple[float, ...]:
         b = self.bias / 100
         b = max(0.1, min(b, 0.9))
         return tuple(repeat(b / self.full_size, self.full_size)) + (1.0 - b,)
 
 
+def set_bias(biases: Sequence[float], idx: int, target: float) -> list[float]:
+    remainder = 1 - target
+    previous_remainder = sum(x for i, x in enumerate(biases) if i != idx)
+    ans = [1. for i in range(len(biases))]
+    for i in range(len(biases)):
+        if i == idx:
+            ans[i] = target
+        else:
+            ans[i] = remainder * biases[i] / previous_remainder
+    return ans
+
+
 class Tall(Layout):
 
     name = 'tall'
@@ -90,15 +113,26 @@ class Tall(Layout):
         return self.layout_opts.full_size
 
     def remove_all_biases(self) -> bool:
-        self.main_bias: List[float] = list(self.layout_opts.build_bias_list())
-        self.biased_map: Dict[int, float] = {}
+        self.main_bias: list[float] = list(self.layout_opts.build_bias_list())
+        self.biased_map: dict[int, float] = {}
         return True
 
-    def variable_layout(self, all_windows: WindowList, biased_map: Dict[int, float]) -> LayoutDimension:
+    def variable_layout(self, all_windows: WindowList, biased_map: dict[int, float]) -> LayoutDimension:
         num = all_windows.num_groups - self.num_full_size_windows
         bias = biased_map if num > 1 else None
         return self.perp_axis_layout(all_windows.iter_all_layoutable_groups(), bias=bias, offset=self.num_full_size_windows)
 
+    def bias_slot(self, all_windows: WindowList, idx: int, fractional_bias: float, cell_increment_bias_h: float, cell_increment_bias_v: float) -> bool:
+        if idx < len(self.main_bias):
+            before_main_bias = self.main_bias
+            self.main_bias = set_bias(self.main_bias, idx, fractional_bias)
+            return self.main_bias != before_main_bias
+
+        before_layout = tuple(self.variable_layout(all_windows, self.biased_map))
+        self.biased_map[idx - self.num_full_size_windows] = cell_increment_bias_v if self.main_is_horizontal else cell_increment_bias_h
+        after_layout = tuple(self.variable_layout(all_windows, self.biased_map))
+        return before_layout == after_layout
+
     def apply_bias(self, idx: int, increment: float, all_windows: WindowList, is_horizontal: bool = True) -> bool:
         num_windows = all_windows.num_groups
         if self.main_is_horizontal == is_horizontal:
@@ -115,16 +149,16 @@ class Tall(Layout):
         if idx < self.num_full_size_windows or num_of_short_windows < 2:
             return False
         idx -= self.num_full_size_windows
-        before_layout = list(self.variable_layout(all_windows, self.biased_map))
+        before_layout = tuple(self.variable_layout(all_windows, self.biased_map))
         before = self.biased_map.get(idx, 0.)
         candidate = self.biased_map.copy()
         candidate[idx] = after = before + increment
-        if before_layout == list(self.variable_layout(all_windows, candidate)):
+        if before_layout == tuple(self.variable_layout(all_windows, candidate)):
             return False
         self.biased_map = candidate
         return before != after
 
-    def simple_layout(self, all_windows: WindowList) -> Generator[Tuple[WindowGroup, LayoutData, LayoutData, bool], None, None]:
+    def simple_layout(self, all_windows: WindowList) -> Generator[tuple[WindowGroup, LayoutData, LayoutData, bool], None, None]:
         num = all_windows.num_groups
         is_fat = not self.main_is_horizontal
         mirrored = self.layout_opts.mirrored
@@ -140,7 +174,7 @@ class Tall(Layout):
                 xl, yl = yl, xl
             yield wg, xl, yl, True
 
-    def full_layout(self, all_windows: WindowList) -> Generator[Tuple[WindowGroup, LayoutData, LayoutData, bool], None, None]:
+    def full_layout(self, all_windows: WindowList) -> Generator[tuple[WindowGroup, LayoutData, LayoutData, bool], None, None]:
         is_fat = not self.main_is_horizontal
         mirrored = self.layout_opts.mirrored
         groups = tuple(all_windows.iter_all_layoutable_groups())
@@ -195,7 +229,7 @@ class Tall(Layout):
     def neighbors_for_window(self, window: WindowType, windows: WindowList) -> NeighborsMap:
         return neighbors_for_tall_window(self.num_full_size_windows, window, windows, self.layout_opts.mirrored, self.main_is_horizontal)
 
-    def layout_action(self, action_name: str, args: Sequence[str], all_windows: WindowList) -> Optional[bool]:
+    def layout_action(self, action_name: str, args: Sequence[str], all_windows: WindowList) -> bool | None:
         if action_name == 'increase_num_full_size_windows':
             self.layout_opts.full_size += 1
             self.main_bias = list(self.layout_opts.build_bias_list())
@@ -249,8 +283,8 @@ class Tall(Layout):
             layout = (x[:3] for x in self.simple_layout(all_windows))
             yield from borders(layout, self.main_is_horizontal, all_windows)
             return
-        main_layouts: List[Tuple[WindowGroup, LayoutData, LayoutData]] = []
-        perp_borders: List[BorderLine] = []
+        main_layouts: list[tuple[WindowGroup, LayoutData, LayoutData]] = []
+        perp_borders: list[BorderLine] = []
         layouts = (self.simple_layout if num <= self.num_full_size_windows else self.full_layout)(all_windows)
         needs_borders_map = all_windows.compute_needs_borders_map(lgd.draw_active_borders)
         active_group = all_windows.active_group
@@ -311,7 +345,7 @@ class Tall(Layout):
         )
         yield from perp_borders[1:-1]
 
-    def layout_state(self) -> Dict[str, Any]:
+    def layout_state(self) -> dict[str, Any]:
         return {
             'num_full_size_windows': self.num_full_size_windows,
             'main_bias': self.main_bias,
