@@ -318,17 +318,36 @@ def remote_control_script(func: str, rest: str) -> FuncArgsType:
     return func, args
 
 
-@func_with_args('nth_os_window', 'nth_window', 'scroll_to_prompt', 'visual_window_select_action_trigger', 'next_layout')
+@func_with_args('nth_os_window', 'nth_window', 'visual_window_select_action_trigger', 'next_layout')
 def single_integer_arg(func: str, rest: str) -> FuncArgsType:
     try:
         num = int(rest)
     except Exception:
         if rest:
             log_error(f'Invalid number for {func}: {rest}')
-        num = -1 if func == 'scroll_to_prompt' else 1
+        num = 1
     return func, [num]
 
 
+@func_with_args('scroll_to_prompt')
+def scroll_to_prompt(func: str, rest: str) -> FuncArgsType:
+    vals = rest.strip().split()
+    args = [-1, 0]
+    if len(vals) > 2:
+        log_error('scroll_to_prompt needs one or two arguments, using defaults')
+    else:
+        try:
+            args[0] = int(vals[0])
+        except Exception:
+            log_error(f'{vals[0]} is not a valid number of prompts to jump for scroll_to_prompt')
+        if len(vals) == 2:
+            try:
+                args[1] = int(vals[1])
+            except Exception:
+                log_error(f'{vals[1]} is not a valid scroll offset for scroll_to_prompt')
+    return func, args
+
+
 @func_with_args('sleep')
 def sleep(func: str, sleep_time: str) -> FuncArgsType:
     mult = 1
@@ -1031,7 +1050,7 @@ def underline_exclusion(x: str) -> tuple[float, Literal['', 'px', 'pt']]:
         try:
             val = float(x[:-2])
         except Exception:
-            raise ValueError(f'Invalid underline_exclusion with non numberic value: {x}')
+            raise ValueError(f'Invalid underline_exclusion with non numeric value: {x}')
         return val, unit
 
 
