@@ -480,6 +480,7 @@ class Tab:  # {{{
         hold: bool = False,
         pass_fds: tuple[int, ...] = (),
         remote_control_fd: int = -1,
+        hold_after_ssh: bool = False
     ) -> Child:
         check_for_suitability = True
         if cmd is None:
@@ -529,7 +530,7 @@ class Tab:  # {{{
             fenv['WINDOWID'] = str(pwid)
         ans = Child(
                 cmd, cwd or self.cwd, stdin, fenv, cwd_from, is_clone_launch=is_clone_launch,
-                add_listen_on_env_var=add_listen_on_env_var, hold=hold, pass_fds=pass_fds, remote_control_fd=remote_control_fd)
+                add_listen_on_env_var=add_listen_on_env_var, hold=hold, pass_fds=pass_fds, remote_control_fd=remote_control_fd, hold_after_ssh=hold_after_ssh)
         ans.fork()
         return ans
 
@@ -568,11 +569,12 @@ class Tab:  # {{{
         pass_fds: tuple[int, ...] = (),
         remote_control_fd: int = -1,
         next_to: Window | None = None,
+        hold_after_ssh: bool = False
     ) -> Window:
         child = self.launch_child(
             use_shell=use_shell, cmd=cmd, stdin=stdin, cwd_from=cwd_from, cwd=cwd, env=env,
             is_clone_launch=is_clone_launch, add_listen_on_env_var=False if allow_remote_control and remote_control_passwords else True,
-            hold=hold, pass_fds=pass_fds, remote_control_fd=remote_control_fd,
+            hold=hold, pass_fds=pass_fds, remote_control_fd=remote_control_fd, hold_after_ssh=hold_after_ssh
         )
         window = Window(
             self, child, self.args, override_title=override_title,
@@ -1215,6 +1217,10 @@ class TabManager:  # {{{
                     return idx
             return -1
 
+        def remove_from_end_of_active_history(idx: int) -> None:
+            while self.active_tab_history and idx_for_id(self.active_tab_history[-1]) == idx:
+                self.active_tab_history.pop()
+
         if active_tab_needs_to_change:
             next_active_tab = -1
             tss = get_options().tab_switch_strategy
@@ -1224,10 +1230,13 @@ class TabManager:  # {{{
                     next_active_tab = idx_for_id(tab_id)
             elif tss == 'left':
                 next_active_tab = max(0, self.active_tab_idx - 1)
+                remove_from_end_of_active_history(next_active_tab)
             elif tss == 'right':
                 next_active_tab = min(self.active_tab_idx, len(self.tabs) - 1)
+                remove_from_end_of_active_history(next_active_tab)
             elif tss == 'last':
                 next_active_tab = len(self.tabs) - 1
+                remove_from_end_of_active_history(next_active_tab)
 
             if next_active_tab < 0:
                 next_active_tab = max(0, min(self.active_tab_idx, len(self.tabs) - 1))
